"""Reader function for KEPSEISMIC community light curve products."""
from ..lightcurve import KeplerLightCurve

from .generic import read_generic_lightcurve

def read_kepseismic_lightcurve(filename, **kwargs):
    """Read a KEPSEISMIC light curve file.

    More information: https://archive.stsci.edu/prepds/kepseismic

    Parameters
    ----------
    filename : str
        Path or URL of a K2SFF light curve FITS file.

    Returns
    -------
    lc : `KeplerLightCurve`
        A populated light curve object.
    """

    lc = read_generic_lightcurve(
        filename,
        time_format='bkjd')

    lc.meta["AUTHOR"] = "KEPSEISMIC"
    lc.meta["TARGETID"] = lc.meta.get("KEPLERID")

    # KEPSEISMIC light curves are normalized by default
    lc.meta["NORMALIZED"] = True

    return KeplerLightCurve(data=lc, **kwargs)



"""Reader for official Kepler light curve FITS files produced by the Ames pipeline."""
from ..lightcurve import KeplerLightCurve
from ..utils import KeplerQualityFlags

from .generic import read_generic_lightcurve


def read_kepler_lightcurve(
    filename, flux_column="pdcsap_flux", quality_bitmask="default"
):
    """Returns a KeplerLightCurve.

    Parameters
    ----------
    filename : str
        Local path or remote url of a Kepler light curve FITS file.
    flux_column : 'pdcsap_flux' or 'sap_flux'
        Which column in the FITS file contains the preferred flux data?
    quality_bitmask : str or int
        Bitmask (integer) which identifies the quality flag bitmask that should
        be used to mask out bad cadences. If a string is passed, it has the
        following meaning:

            * "none": no cadences will be ignored (`quality_bitmask=0`).
            * "default": cadences with severe quality issues will be ignored
              (`quality_bitmask=1130799`).
            * "hard": more conservative choice of flags to ignore
              (`quality_bitmask=1664431`). This is known to remove good data.
            * "hardest": removes all data that has been flagged
              (`quality_bitmask=2096639`). This mask is not recommended.

        See the :class:`KeplerQualityFlags` class for details on the bitmasks.
    """
    lc = read_generic_lightcurve(
        filename,
        flux_column=flux_column,
        quality_column="sap_quality",
        time_format="bkjd",
    )

    # Filter out poor-quality data
    # NOTE: Unfortunately Astropy Table masking does not yet work for columns
    # that are Quantity objects, so for now we remove poor-quality data instead
    # of masking. Details: https://github.com/astropy/astropy/issues/10119
    quality_mask = KeplerQualityFlags.create_quality_mask(
        quality_array=lc["sap_quality"], bitmask=quality_bitmask
    )
    lc = lc[quality_mask]

    lc.meta["AUTHOR"] = "Kepler"
    lc.meta["TARGETID"] = lc.meta.get("KEPLERID")
    lc.meta["QUALITY_BITMASK"] = quality_bitmask
    lc.meta["QUALITY_MASK"] = quality_mask
    return KeplerLightCurve(data=lc)



"""Read a generic FITS table containing a light curve."""
import logging

from astropy.io import fits
from astropy.table import Table
from astropy.time import Time
import numpy as np

from ..utils import validate_method
from ..lightcurve import LightCurve
from ..units import ppm


log = logging.getLogger(__name__)


def read_generic_lightcurve(
    filename,
    time_column="time",
    flux_column="flux",
    flux_err_column="flux_err",
    quality_column="quality",
    cadenceno_column="cadenceno",
    centroid_col_column="mom_centr1",
    centroid_row_column="mom_centr2",
    time_format=None,
    ext=1,
):
    """Generic helper function to convert a Kepler ot TESS light curve file
    into a generic `LightCurve` object.
    """
    if isinstance(filename, fits.HDUList):
        hdulist = filename  # Allow HDUList to be passed
    else:
        hdulist = fits.open(filename)

    # Raise an exception if the requested extension is invalid
    if isinstance(ext, str):
        validate_method(ext, supported_methods=[hdu.name.lower() for hdu in hdulist])
    tab = Table.read(hdulist[ext], format="fits")

    # Make sure the meta data also includes header fields from extension #0
    tab.meta.update(hdulist[0].header)

    tab.meta = {k: v for k, v in tab.meta.items()}

    for colname in tab.colnames:
        # Ensure units have the correct astropy format
        # Speed-up: comparing units by their string representation is 1000x
        # faster than performing full-blown unit comparison
        unitstr = str(tab[colname].unit)
        if unitstr == "e-/s":
            tab[colname].unit = "electron/s"
        elif unitstr == "pixels":
            tab[colname].unit = "pixel"
        elif unitstr == "ppm" and repr(tab[colname].unit).startswith("Unrecognized"):
            # Workaround for issue #956
            tab[colname].unit = ppm

        # Rename columns to lowercase
        tab.rename_column(colname, colname.lower())

    # Some KEPLER files used to have a T column instead of TIME.
    if time_column == "time" and "time" not in tab.columns and "t" in tab.colnames:
        tab.rename_column("t", "time")
    if time_column != "time":
        tab.rename_column(time_column, "time")

    # We *have* to remove rows with TIME=NaN because the Astropy Time
    # object does not support the presence of NaNs.
    # Fortunately, such rows are always bad data.
    nans = np.isnan(tab["time"].data)
    if np.any(nans):
        log.debug("Ignoring {} rows with NaN times".format(np.sum(nans)))
    tab = tab[~nans]

    # Prepare a special time column
    if not time_format:
        if hdulist[ext].header.get("BJDREFI") == 2454833:
            time_format = "bkjd"
        elif hdulist[ext].header.get("BJDREFI") == 2457000:
            time_format = "btjd"
        else:
            raise ValueError(f"Input file has unclear time format: {filename}")
    time = Time(
        tab["time"].data,
        scale=hdulist[ext].header.get("TIMESYS", "tdb").lower(),
        format=time_format,
    )
    tab.remove_column("time")

    # For backwards compatibility with Lightkurve v1.x,
    # we make sure standard columns and attributes exist.
    if "flux" not in tab.columns:
        tab.add_column(tab[flux_column], name="flux", index=0)
    if "flux_err" not in tab.columns:
        # Try falling back to `{flux_column}_err` if possible
        if flux_err_column not in tab.columns:
            flux_err_column = flux_column + "_err"
        if flux_err_column in tab.columns:
            tab.add_column(tab[flux_err_column], name="flux_err", index=1)
    if "quality" not in tab.columns and quality_column in tab.columns:
        tab.add_column(tab[quality_column], name="quality", index=2)
    if "cadenceno" not in tab.columns and cadenceno_column in tab.columns:
        tab.add_column(tab[cadenceno_column], name="cadenceno", index=3)
    if "centroid_col" not in tab.columns and centroid_col_column in tab.columns:
        tab.add_column(tab[centroid_col_column], name="centroid_col", index=4)
    if "centroid_row" not in tab.columns and centroid_row_column in tab.columns:
        tab.add_column(tab[centroid_row_column], name="centroid_row", index=5)

    tab.meta["LABEL"] = hdulist[0].header.get("OBJECT")
    tab.meta["MISSION"] = hdulist[0].header.get(
        "MISSION", hdulist[0].header.get("TELESCOP")
    )
    tab.meta["RA"] = hdulist[0].header.get("RA_OBJ")
    tab.meta["DEC"] = hdulist[0].header.get("DEC_OBJ")
    tab.meta["FILENAME"] = filename
    tab.meta["FLUX_ORIGIN"] = flux_column

    return LightCurve(time=time, data=tab)



"""Define custom AstroPy units commonly used by the Kepler/TESS community."""
from astropy import units as u

__all__ = ["ppt", "ppm"]

ppt = u.def_unit(["ppt", "parts per thousand"], u.Unit(1e-3))
ppm = u.def_unit(["ppm", "parts per million"], u.Unit(1e-6))
u.add_enabled_units([ppt, ppm])



"""Defines collections of data products."""
import warnings

import matplotlib
import matplotlib.pyplot as plt
import numpy as np

from astropy.table import vstack
from astropy.utils.decorators import deprecated

from . import MPLSTYLE
from .targetpixelfile import TargetPixelFile
from .utils import LightkurveWarning, LightkurveDeprecationWarning


__all__ = ["LightCurveCollection", "TargetPixelFileCollection"]


class Collection(object):
    """Base class for `LightCurveCollection` and `TargetPixelFileCollection`.

    A collection can be indexed by standard Python list syntax.
    Additionally, it can be indexed by a subset of `numpy.ndarray` syntax:
    boolean array indexing and integer array indexing.

    Attributes
    ----------
    data: array-like
        List of data objects.

    Examples
    --------
    Filter a collection by boolean array indexing.

        >>> lcc_filtered = lcc[(lcc.sector >= 13) & (lcc.sector <= 19)]  # doctest: +SKIP
        >>> lc22 = lcc[lcc.sector == 22][0]  # doctest: +SKIP

    Filter a collection by integer array indexing to get the object at index 0 and 2.

        >>>  lcc_filtered = lcc[0, 2]  # doctest: +SKIP
    """

    def __init__(self, data):
        if data is not None:
            # ensure we have our own container
            self.data = [item for item in data]
        else:
            self.data = []

    def __len__(self):
        return len(self.data)

    def __getitem__(self, index_or_mask):
        if isinstance(index_or_mask, (int, np.integer)):
            return self.data[index_or_mask]
        elif isinstance(index_or_mask, slice):
            return type(self)(self.data[index_or_mask])
        elif all([isinstance(i, (bool, np.bool_)) for i in index_or_mask]):
            # case indexOrMask is bool array like, e.g., np.ndarray, collections.abc.Sequence, etc.

            # note: filter using nd.array is very slow
            #   np.array(self.data)[np.nonzero(indexOrMask)]
            # specifically, nd.array(self.data) is very slow, as it deep copies the data
            # so we create the filtered list on our own
            if len(index_or_mask) != len(self.data):
                raise IndexError(
                    f"boolean index did not match indexed array; dimension is {len(self.data)} "
                    f"but corresponding boolean dimension is {len(index_or_mask)}"
                )
            return type(self)([self.data[i] for i in np.nonzero(index_or_mask)[0]])
        elif all([isinstance(i, (int, np.integer)) for i in index_or_mask]):
            # case int array like, follow ndarray behavior
            return type(self)([self.data[i] for i in index_or_mask])
        else:
            raise IndexError(
                "only integers, slices (`:`) and integer or boolean arrays are valid indices"
            )

    def __setitem__(self, index, obj):
        self.data[index] = obj

    def append(self, obj):
        """Appends a new object to the collection.

        Parameters
        ----------
        obj : object
            Typically a LightCurve or TargetPixelFile object
        """
        self.data.append(obj)

    def __repr__(self):
        result = f"{self.__class__.__name__} of {len(self)} objects:\n    "
        # LightCurve objects provide a special `_repr_simple_` method
        # to avoid printing an entire table here
        result += "\n    ".join(
            [
                f"{idx}: " + getattr(obj, "_repr_simple_", obj.__repr__)()
                for idx, obj in enumerate(self)
            ]
        )
        return result

    def _safeGetScalarAttr(self, attrName):
        # return np.nan when the attribute is missing, so that the returned value can be used in a comparison
        # e.g., lcc[lcc.sector < 25]
        return np.array([getattr(lcOrTpf, attrName, np.nan) for lcOrTpf in self.data])

    @property
    def sector(self):
        """The sector number for TESS data products.

        Returns `numpy.nan` for data products with lack a sector meta data keyword.
        The attribute is useful for filtering a collection by sector.

        Examples
        --------
        Plot two lightcurves, one from TESS sectors 13 to 19, and one for sector 22.

            >>> import lightkurve as lk
            >>> lcc = lk.search_lightcurve('TIC286923464', author='SPOC').download_all()  # doctest: +SKIP
            >>> lcc_filtered = lcc[(lcc.sector >= 13) & (lcc.sector <= 19)]  # doctest: +SKIP
            >>> lcc_filtered.plot()  # doctest: +SKIP
            >>> lcc[lcc.sector == 22][0].plot()  # doctest: +SKIP

        """
        return self._safeGetScalarAttr("sector")

    @property
    def quarter(self):
        """(Kepler-specific) the quarters of the lightcurves / target pixel files.

        The Kepler quarters of the lightcurves / target pixel files; `numpy.nan` for those with none.
        """
        return self._safeGetScalarAttr("quarter")

    @property
    def campaign(self):
        """(K2-specific) the campaigns of the lightcurves / target pixel files.

        The K2 campaigns of the lightcurves / target pixel files; `numpy.nan` for those with none.
        """
        return self._safeGetScalarAttr("campaign")


class LightCurveCollection(Collection):
    """Class to hold a collection of LightCurve objects.

    Attributes
    ----------
    lightcurves : array-like
        List of LightCurve objects.
    """

    def __init__(self, lightcurves):
        super(LightCurveCollection, self).__init__(lightcurves)

    @property
    @deprecated("2.0", warning_type=LightkurveDeprecationWarning)
    def PDCSAP_FLUX(self):
        """DEPRECATED. Replaces `LightCurveFileCollection.PDCSAP_FLUX`.
        Provided for backwards-compatibility with Lightkurve v1.x;
        will be removed soon."""
        return LightCurveCollection([lc.PDCSAP_FLUX for lc in self])

    @property
    @deprecated("2.0", warning_type=LightkurveDeprecationWarning)
    def SAP_FLUX(self):
        """DEPRECATED. Replaces `LightCurveFileCollection.SAP_FLUX`.
        Provided for backwards-compatibility with Lightkurve v1.x;
        will be removed soon."""
        return LightCurveCollection([lc.SAP_FLUX for lc in self])

    def stitch(self, corrector_func=lambda x: x.normalize()):
        """Stitch all light curves in the collection into a single `LightCurve`.

        Any function passed to `corrector_func` will be applied to each light curve
        before stitching. For example, passing "lambda x: x.normalize().flatten()"
        will normalize and flatten each light curve before stitching.

        Parameters
        ----------
        corrector_func : function
            Function that accepts and returns a `~lightkurve.lightcurve.LightCurve`.
            This function is applied to each light curve in the collection
            prior to stitching. The default is to normalize each light curve.

        Returns
        -------
        lc : `~lightkurve.lightcurve.LightCurve`
            Stitched light curve.
        """
        if corrector_func is None:
            corrector_func = lambda x: x  # noqa: E731
        with warnings.catch_warnings():  # ignore "already normalized" message
            warnings.filterwarnings("ignore", message=".*already.*")
            lcs = [corrector_func(lc) for lc in self]

        # Address issue #954: ignore incompatible columns with the same name
        columns_to_remove = set()
        for col in lcs[0].columns:
            for lc in lcs[1:]:
                if col in lc.columns:
                    if not (
                        issubclass(lcs[0][col].__class__, lc[col].__class__)
                        or lcs[0][col].__class__.info is lc[col].__class__.info
                    ):
                        columns_to_remove.add(col)
                        continue

        if len(columns_to_remove) > 0:
            warnings.warn(
                f"The following columns will be excluded from stitching because the column types are incompatible: {columns_to_remove}",
                LightkurveWarning,
            )
            lcs = [lc.copy() for lc in lcs]
            [
                lc.remove_columns(columns_to_remove.intersection(lc.columns))
                for lc in lcs
            ]

        # Need `join_type='inner'` until AstroPy supports masked Quantities
        return vstack(lcs, join_type="inner", metadata_conflicts="silent")

    def plot(self, ax=None, offset=0.0, **kwargs) -> matplotlib.axes.Axes:
        """Plots all light curves in the collection on a single plot.

        Parameters
        ----------
        ax : `~matplotlib.axes.Axes`
            A matplotlib axes object to plot into. If no axes is provided,
            a new one will be created.
        offset : float
            Offset to add to targets with different labels, to prevent light
            curves from being plotted on top of each other.  For example, if
            the collection contains light curves with unique labels "A", "B",
            and "C", light curves "A" will have `0*offset` added to their flux,
            light curves "B" will have `1*offset` offset added, and "C" will
            have `2*offset` added.
        **kwargs : dict
            Dictionary of arguments to be passed to `LightCurve.plot`.

        Returns
        -------
        ax : `~matplotlib.axes.Axes`
            The matplotlib axes object.
        """
        with plt.style.context(MPLSTYLE):
            if ax is None:
                _, ax = plt.subplots()
            for kwarg in ["c", "color", "label"]:
                if kwarg in kwargs:
                    kwargs.pop(kwarg)

            for idx, lc in enumerate(self):
                kwargs["label"] = f"{idx}: {lc.meta.get('LABEL', '(missing label)')}"
                lc.plot(ax=ax, c=f"C{idx}", offset=idx * offset, **kwargs)

            # If some but not all light curves are normalized, ensure the Y label
            # says "Flux" and not "Normalized Flux"
            normstatus = [lc.meta.get("NORMALIZED", False) for lc in self]
            if "normalize" not in kwargs and any(normstatus) and not all(normstatus):
                warnings.warn(
                    "Some but not all of the light curves in the collection appear to be normalized. "
                    "You may wish to use `normalize=True` to ensure all are normalized.",
                    LightkurveWarning,
                )
                if "ylabel" not in kwargs:
                    ax.set_ylabel("Flux")

        return ax


class TargetPixelFileCollection(Collection):
    """Class to hold a collection of `~lightkurve.targetpixelfile.TargetPixelFile` objects.

    Parameters
    ----------
    tpfs : list or iterable
        List of `~lightkurve.targetpixelfile.TargetPixelFile` objects.
    """

    def __init__(self, tpfs):
        super(TargetPixelFileCollection, self).__init__(tpfs)

    def plot(self, ax=None):
        """Individually plots all TargetPixelFile objects in a single
        matplotlib axes object.

        Parameters
        ----------
        ax : `~matplotlib.axes.Axes`
            A matplotlib axes object to plot into. If no axes is provided,
            a new one will be created.

        Returns
        -------
        ax : `~matplotlib.axes.Axes`
            The matplotlib axes object.
        """
        if ax is None:
            _, ax = plt.subplots(len(self.data), 1, figsize=(7, (7 * len(self.data))))
        if len(self.data) == 1:
            self.data[0].plot(ax=ax)
        else:
            for i, tpf in enumerate(self.data):
                tpf.plot(ax=ax[i])
        return ax



"""Functions for reading light curve data."""
import logging

from astropy.io import fits
from astropy.utils import deprecated

from .detect import detect_filetype
from ..lightcurve import KeplerLightCurve, TessLightCurve
from ..utils import LightkurveDeprecationWarning, LightkurveError

log = logging.getLogger(__name__)


__all__ = ["open", "read"]


@deprecated("2.0", alternative="read()", warning_type=LightkurveDeprecationWarning)
def open(path_or_url, **kwargs):
    """DEPRECATED. Please use `lk.read()` instead.

    This function has been deprecated because its name collides with Python's
    built-in `open()` function.
    """
    return read(path_or_url, **kwargs)


def read(path_or_url, **kwargs):
    """Reads any valid Kepler or TESS data file and returns an instance of
    `~lightkurve.lightcurve.LightCurve` or
    :class:`TargetPixelFile <lightkurve.targetpixelfile>`

    This function will automatically detect the type of the data product, and return the
    appropriate object. File types currently supported include::

        * `KeplerTargetPixelFile` (typical suffix "-targ.fits.gz");
        * `KeplerLightCurve` (typical suffix "llc.fits");
        * `TessTargetPixelFile` (typical suffix "_tp.fits");
        * `TessLightCurve` (typical suffix "_lc.fits").

    Parameters
    ----------
    path_or_url : str
        Path or URL of a FITS file.
    quality_bitmask : str or int, optional
        Bitmask (integer) which identifies the quality flag bitmask that should
        be used to mask out bad cadences. If a string is passed, it has the
        following meaning:

            * "none": no cadences will be ignored
            * "default": cadences with severe quality issues will be ignored
            * "hard": more conservative choice of flags to ignore
              This is known to remove good data.
            * "hardest": removes all data that has been flagged
              This mask is not recommended.

        See the :class:`KeplerQualityFlags <lightkurve.utils.KeplerQualityFlags>` or :class:`TessQualityFlags <lightkurve.utils.TessQualityFlags>` class for details on the bitmasks.
    flux_column : str, optional
        (Applicable to LightCurve products only) The column in the FITS file to be read as `flux`. Defaults to 'pdcsap_flux'.
        Typically 'pdcsap_flux' or 'sap_flux'.

    Returns
    -------
    data : a subclass of :class:`TargetPixelFile <lightkurve.targetpixelfile>` or `~lightkurve.lightcurve.LightCurve`
           depending on the detected file type.

    Raises
    ------
    ValueError : raised if the data product is not recognized as a Kepler or TESS product.

    Examples
    --------
    To read a target pixel file using its path or URL, simply use:

        >>> import lightkurve as lk
        >>> tpf = lk.read("mytpf.fits")  # doctest: +SKIP
    """
    log.debug("Opening {}.".format(path_or_url))
    # pass header into `detect_filetype()`
    try:
        with fits.open(path_or_url) as temp:
            filetype = detect_filetype(temp)
            log.debug("Detected filetype: '{}'.".format(filetype))
    except OSError as e:
        filetype = None
        # Raise an explicit FileNotFoundError if file not found
        if "No such file" in str(e):
            raise e

    if filetype == "KeplerLightCurve":
        return KeplerLightCurve.read(path_or_url, format="kepler", **kwargs)
    elif filetype == "TessLightCurve":
        return TessLightCurve.read(path_or_url, format="tess", **kwargs)
    elif filetype == "QLP":
        return TessLightCurve.read(path_or_url, format="qlp", **kwargs)
    elif filetype == "PATHOS":
        return TessLightCurve.read(path_or_url, format="pathos", **kwargs)
    elif filetype == "CDIPS":
        return TessLightCurve.read(path_or_url, format='cdips', **kwargs)
    elif filetype == "TASOC":
        return TessLightCurve.read(path_or_url, format="tasoc", **kwargs)
    elif filetype == "K2SFF":
        return KeplerLightCurve.read(path_or_url, format="k2sff", **kwargs)
    elif filetype == "EVEREST":
        return KeplerLightCurve.read(path_or_url, format="everest", **kwargs)
    elif filetype == "KEPSEISMIC":
        return KeplerLightCurve.read(path_or_url, format="kepseismic", **kwargs)

    # Official data products;
    # if the filetype is recognized, instantiate a class of that name
    if filetype is not None:
        try:
            return getattr(__import__("lightkurve"), filetype)(path_or_url, **kwargs)
        except AttributeError as exc:
            raise LightkurveError(
                f"{filetype} files are not supported " "in this version of Lightkurve."
            ) from exc
    else:
        # if these keywords don't exist, raise `ValueError`
        raise LightkurveError(
            "Not recognized as a supported data product:\n"
            f"{path_or_url}\n"
            "This file may be corrupt due to an interrupted download. "
            "Please remove it from your disk and try again."
        )



"""The .io sub-package provides functions for reading data."""
from .detect import *
from .read import *

from . import kepler, tess, qlp, k2sff, everest, pathos, tasoc, kepseismic
from . import cdips
from .. import LightCurve

from astropy.io import registry


__all__ = ["read", "open"]


# We intend the reader functions to be accessed via `LightCurve.read()`,
# so we add them to the `astropy.io.registry`.
try:
    registry.register_reader("kepler", LightCurve, kepler.read_kepler_lightcurve)
    registry.register_reader("tess", LightCurve, tess.read_tess_lightcurve)
    registry.register_reader("qlp", LightCurve, qlp.read_qlp_lightcurve)
    registry.register_reader("k2sff", LightCurve, k2sff.read_k2sff_lightcurve)
    registry.register_reader("everest", LightCurve, everest.read_everest_lightcurve)
    registry.register_reader("pathos", LightCurve, pathos.read_pathos_lightcurve)
    registry.register_reader('cdips', LightCurve, cdips.read_cdips_lightcurve)
    registry.register_reader("tasoc", LightCurve, tasoc.read_tasoc_lightcurve)
    registry.register_reader("kepseismic", LightCurve, kepseismic.read_kepseismic_lightcurve)
except registry.IORegistryError:
    pass  # necessary to enable autoreload during debugging



"""Reader for A PSF-Based Approach to TESS Cluster Difference Imaging Photometric Survey (CDIPS) light curves

Website: https://archive.stsci.edu/hlsp/cdips
Product Description: https://archive.stsci.edu/hlsps/cdips/hlsp_cdips_tess_ffi_all_tess_v01_readme.md
"""
import logging

from ..lightcurve import TessLightCurve
from ..utils import TessQualityFlags

from .generic import read_generic_lightcurve

log = logging.getLogger(__name__)

def read_cdips_lightcurve(filename,
                            flux_column="IRM1",
                            include_inst_errs=False,
                            quality_bitmask=None):
    """Returns a `TessLightCurve`.

    Note: CDIPS light curves have already had quality filtering applied, and
    do not provide the bitflags necessary for a user to apply a new bitmask.
    Therefore, frames corresponding to "momentum dumps and coarse point modes"
    are removed according to Bouma et al. 2019, and no other quality filtering
    is allowed. The `quality_bitmask` parameter is ignored but accepted for
    compatibility with other data format readers.

    Parameters
    ----------
    filename : str
        Local path or remote url of CDIPS light curve FITS file.
    flux_column : str
        'IFL#', 'IRM#', 'TFA#', or 'PCA#' (# = 1, 2, or 3)
        Which column in the FITS file contains the preferred flux data?
    include_inst_errs: bool
        Whether to include the instrumental flux/magnitude errors
        (Errors are not provided for trend-filtered magnitudes)
    """
    ap = flux_column[-1]

    # Only the instrumental magnitudes are provided, and are not provided for
    # trend-filtered light curves. User should select whether to include the
    # instrumental errors or ignore them
    if include_inst_errs:
        # If fluxes are requested, return flux errors
        if flux_column[:-1].lower()=="ifl":
            flux_err_column = f"ife{ap}"
        # Otherwise magnitudes are being requested, return magnitude errors
        else:
            flux_err_column = f"ire{ap}"
    else:
        flux_err_column = ""

    # Set the appropriate error column for this aperture
    quality_column = f"irq{ap}"

    lc = read_generic_lightcurve(filename,
                                 time_column="tmid_bjd",
                                 flux_column=flux_column.lower(),
                                 flux_err_column=flux_err_column,
                                 quality_column=quality_column,
                                 time_format='btjd')

    # Filter out poor-quality data
    # NOTE: Unfortunately Astropy Table masking does not yet work for columns
    # that are Quantity objects, so for now we remove poor-quality data instead
    # of masking. Details: https://github.com/astropy/astropy/issues/10119

    # CDIPS uses their own quality keywords instead of the default bitflags
    # Based on Bouma+2019, they filter out coarse point (4) and desat (32)
    # as well as other cadences flagged for particular sectors
    quality_mask = (lc['quality']=="G") | (lc['quality']=="0")
    lc = lc[quality_mask]

    lc.meta["AUTHOR"] = "CDIPS"
    lc.meta['TARGETID'] = lc.meta.get('TICID')
    lc.meta['QUALITY_BITMASK'] = 36
    lc.meta['QUALITY_MASK'] = quality_mask

    return TessLightCurve(data=lc)



"""Metrics to assess under- and over-fitting of systematic noise.

This module provides two metrics, `overfit_metric_lombscargle` and `underfit_metric_neighbors`,
which enable users to assess whether the noise in a systematics-corrected light curve has been
under- or over-fitted.  These features were contributed by Jeff Smith (cf. https://github.com/lightkurve/lightkurve/pull/855)
and are in turn inspired by similar metrics in use by the PDC module of the official Kepler/TESS pipeline.
"""
import logging

import numpy as np
from scipy.interpolate import PchipInterpolator
from memoization import cached

from .. import LightCurve


log = logging.getLogger(__name__)


def overfit_metric_lombscargle(
    original_lc: LightCurve, corrected_lc: LightCurve, n_samples: int = 10
) -> float:
    """Uses a LombScarglePeriodogram to assess the change in broad-band
    power in a corrected light curve to measure the degree of over-fitting.

    The to_periodogram Lomb-Scargle method is used and the sampling band is
    from one frequency separation to the Nyquist frequency

    This over-fitting goodness metric is calibrated such that a metric
    value of 0.5 means the introduced noise due to over-fitting is at the
    same power level as the uncertainties in the light curve.

    Parameters
    ----------
    original_lc : LightCurve
        Uncorrected light curve.
    corrected_lc : LightCurve
        Light curve from which systematics have been removed.
    n_samples : int
        The number of times to compute and average the metric
        This can stabilize the value, default = 10

    Returns
    -------
    overfit_metric : float
        A float in the range [0,1] where 0 => Bad, 1 => Good
    """
    # The fit can sometimes result in NaNs
    # Also median normalize original and correctod LCs
    orig_lc = original_lc.copy()
    orig_lc = orig_lc.remove_nans().normalize()
    orig_lc -= 1.0
    corrected_lc = corrected_lc.copy()
    corrected_lc = corrected_lc.remove_nans().normalize()
    corrected_lc -= 1.0
    if len(corrected_lc) == 0:
        return 1.0

    # Perform the measurement multiple times and average to stabilize the metric
    metric_per_iter = []
    for idx in np.arange(n_samples):
        pgOrig = orig_lc.to_periodogram()
        # Use the same periods in the corrected flux as just used in the
        # original flux
        pgCorrected = corrected_lc.to_periodogram(frequency=pgOrig.frequency)

        # Get an estimate of the PSD at the uncertainties limit
        # The raw and corrected uncertainties should be essentially identical so
        # use the corrected
        # TODO: the periodogram of WGN should be analytical to compute!
        nNonGappedCadences = len(orig_lc)
        meanCorrectedUncertainties = np.nanmean(corrected_lc.flux_err)
        WGNCorrectedUncert = (
            np.random.randn(nNonGappedCadences, 1) * meanCorrectedUncertainties
        ).T[0]
        model_err = np.zeros(nNonGappedCadences)
        noise_lc = LightCurve(
            time=orig_lc.time, flux=WGNCorrectedUncert, flux_err=model_err
        )
        pgCorrectedUncert = noise_lc.to_periodogram()
        meanCorrectedUncertPower = np.nanmean(np.array(pgCorrectedUncert.power))

        # Compute the change in power
        pgChange = np.array(pgCorrected.power) - np.array(pgOrig.power)

        # Ignore nans
        pgChange = pgChange[~np.isnan(pgChange)]

        # If no increase in power in ANY bands then return a perfect loss
        # function
        if len(np.nonzero(pgChange > 0.0)[0]) == 0:
            metric_per_iter.append(0.0)
        else:
            # We are only concerned with bands where the power increased so
            # when(pgCorrected - pgOrig) > 0
            # Normalize by the noise in the uncertainty
            # We want the goodness to begin to degrade when the introduced
            # noise is greater than the uncertainties.
            # So, when Sigmoid > 0.5 (given twiceSigmoidInv defn.)
            denominator = (
                len(np.nonzero(pgChange > 0.0)[0])
            ) * meanCorrectedUncertPower
            if denominator == 0:
                # Suppress divide by zero warning
                result = np.inf
            else:
                result = np.sum(pgChange[pgChange > 0.0]) / denominator
            metric_per_iter.append(result)

    metric = np.mean(metric_per_iter)

    # We want the goodness to span (0,1]
    # Use twice a reversed sigmoid to get a [0,1] range mapped from a [0,inf) range
    def sigmoidInv(x):
        return 2.0 / (1 + np.exp(x))

    # Make sure maximum score is 1.0
    metric = sigmoidInv(np.max([metric, 0.0]))

    return metric


def underfit_metric_neighbors(
    corrected_lc: LightCurve,
    radius: float = 6000,
    min_targets: int = 30,
    max_targets: int = 50,
    interpolate: bool = False,
    extrapolate: bool = False,
):
    """This goodness metric measures the degree of under-fitting of the
    CBVs to the light curve. It does so by measuring the mean residual target to
    target Pearson correlation between the target under study and a selection of
    neighboring SPOC SAP target light curves.

    This function will search within the given radiu in arceseconds and find the
    min_targets nearest targets up until max_targets is reached. If less than
    min_targets is found a MinTargetsError Exception is raised.

    The downloaded neighboring targets will normally be "aligned" to the
    corrected_lc, meaning the cadence numbers are used to align the targets
    to the corrected_lc. However, if interpolate=True then the targets will be
    interpolated to the corrected_lc cadence times. extrapolate=True will 
    further extrapolate the targets to the corrected_lc cadence times.

    The returned under-fitting goodness metric is callibrated such that a
    value of 0.95 means the residual correlations in the target is
    equivalent to chance correlations of White Gaussian Noise.

    Parameters
    ----------
    corrected_lc : LightCurve
        Light curve from which systematics have been removed.
    radius : float
        Search radius to find neighboring targets in arcseconds
    min_targets : int
        Minimum number of targets to use in correlation metric
        Using too few can cause unreliable results. Default = 30
    max_targets : int
        Maximum number of targets to use in correlation metric
        Using too many can slow down the metric due to large data
        download. Default = 50
    interpolate : bool
        If `True`, the flux values of the neighboring light curves will be
        interpolated to match the times of the `corrected_lc`.
        If `False`, the flux values will simply be aligned by time where possible.

    Returns
    -------
    under_fitting_metric : float
        A float in the range [0,1] where 0 => Bad, 1 => Good
    """
    # Download and pre-process neighboring light curves
    lc_neighborhood, lc_neighborhood_flux = _download_and_preprocess_neighbors(
        corrected_lc=corrected_lc,
        radius=radius,
        min_targets=min_targets,
        max_targets=max_targets,
        interpolate=interpolate,
        extrapolate=extrapolate,
        flux_column="sap_flux",
    )

    # If there happens to be any cadences in the corrected_lc
    # that are not in the neighboring targets then those need to
    # be removed.
    # If we interpolated the CBVs then this should not occur
    # Also normalize
    corrected_lc = corrected_lc.copy().remove_nans().normalize()
    corrected_lc -= 1.0
    if interpolate:
        corrected_lc_flux_trimmed = corrected_lc.flux.value
    else:
        corrected_lc_trim_mask = np.in1d(corrected_lc.cadenceno, 
                lc_neighborhood[0].cadenceno)
        corrected_lc_flux_trimmed = corrected_lc.flux.value[corrected_lc_trim_mask]

    # Create fluxMatrix. The last entry is the target under study
    fluxMatrix = np.zeros((len(lc_neighborhood_flux[0]), len(lc_neighborhood_flux) + 1))
    for idx in np.arange(len(fluxMatrix[0, :]) - 1):
        fluxMatrix[:, idx] = lc_neighborhood_flux[idx]
    # Add in the trimmed target under study
    fluxMatrix[:, -1] = corrected_lc_flux_trimmed

    # Ignore NaNs
    mask = ~np.isnan(corrected_lc_flux_trimmed)
    fluxMatrix = fluxMatrix[mask, :]

    # Determine the target-target correlation between target and
    # neighborhood
    correlationMatrix = _compute_correlation(fluxMatrix)

    # The selection basis for targets used for the PDC-MAP SVD  uses median
    # absolute correlation per star.  However, here we wish to overemphasize
    # any residual correlation between a handfull of targets and not the
    # overall correlation (which should almost always be low).

    # We want a residual correlation larger than random correlations of WGN
    # to mean a meaningful correlation. The median Pearson correlation of
    # WGN of nCadences is approximated by the equation:
    # 0.0010288 + 0.80304 nCadences^ -0.50128
    nCadences = len(fluxMatrix[:, 0])
    beta = [0.0007, 0.8083, -0.5023]
    WGNCorrelation = beta[0] + beta[1] * (nCadences ** (beta[2]))

    # badLimit is the goodness value for WGN correlations
    # I.e. anything above this goodness value is equivalent to random correlations
    # I.e. 0.95 = sigmoidInv(WGNCorr * correlationScale)
    badLimit = 0.95
    correlationScale = 1 / (WGNCorrelation) * np.log((2.0 / badLimit) - 1.0)

    # Over-emphasize any individual correlation groups. Note the power of
    # three after taking the absolute value
    # of the correlation. Also, the mean is used so that outliers are *not* ignored.
    # Zero diagonal elements
    correlationMatrix = np.tril(correlationMatrix, k=-1) + np.triu(
        correlationMatrix, k=+1
    )

    # Add up the correlation over all targets ignoring NaNs (no corrected fit)
    correlation = correlationScale * np.nanmean(np.abs(correlationMatrix) ** 3, axis=0)

    # We only want the last entry, which is for the target under study
    correlation = correlation[-1]

    # We want the goodness to span (0,1]
    # Use twice a reversed sigmoid to get a [0,1] range mapped from a [0,inf) range
    def sigmoidInv(x):
        return 2.0 / (1 + np.exp(x))

    metric = sigmoidInv(correlation)

    return metric


# Custom exception to track when minimum targets is not reached
class MinTargetsError(Exception):
    pass


def _unique_key_for_processing_neighbors(
    corrected_lc: LightCurve,
    radius: float = 6000.0,
    min_targets: int = 30,
    max_targets: int = 50,
    interpolate: bool = False,
    extrapolate: bool = False,
    author: tuple = ("Kepler", "K2", "SPOC"),
    flux_column: str = "sap_flux",
):
    """Returns a unique key that will determine whether a cached version of a
    call to `_download_and_preprocess_neighbors` can be re-used."""
    return f"{corrected_lc.ra}{corrected_lc.dec}{corrected_lc.cadenceno}{radius}{min_targets}{max_targets}{author}{flux_column}{interpolate}"


@cached(custom_key_maker=_unique_key_for_processing_neighbors)
def _download_and_preprocess_neighbors(
    corrected_lc: LightCurve,
    radius: float = 6000.0,
    min_targets: int = 30,
    max_targets: int = 50,
    interpolate: bool = False,
    extrapolate: bool = False,
    author: tuple = ("Kepler", "K2", "SPOC"),
    flux_column: str = "sap_flux",
):
    """Returns a collection of neighboring light curves.

    If less than min_targets a MinTargetsError Exception is raised.

    Parameters
    ----------
    corrected_lc : LightCurve
        Light curve around which to look for neighbors.
    radius : float
        Conesearch radius in arcseconds.
    min_targets : int
        Minimum number of targets to return.
        A `ValueError` will be raised if this number cannot be obtained.
    max_targets : int
        Maximum number of targets to return.
        Using too many can slow down this function due to large data
        download.
    interpolate : bool
        If `True`, the flux values of the neighboring light curves will be
        interpolated to match the times of the `corrected_lc`.
        If `False`, the flux values will simply be aligned by time where possible.
    extrapolate : bool
        If `True`, the  flux values of the neighboring light curves will be
        also be extrapolated. Note: extrapolated values can be unstable.

    Returns
    -------
    lc_neighborhood : LightCurveCollection
        Collection of all neighboring light curves used.
    lc_neighborhood_flux : list
        List containing the flux arrays of the neighboring light curves,
        interpolated or aligned with `corrected_lc` if requested.
    """
    if extrapolate and (extrapolate != interpolate):
        raise Exception('interpolate must be True if extrapolate is True')

    search = corrected_lc.search_neighbors(
        limit=max_targets, radius=radius, author=author
    )
    if len(search) < min_targets:
        raise MinTargetsError(
            f"Unable to find at least {min_targets} neighbors within {radius} arcseconds radius."
        )
    log.info(
        f"Downloading {len(search)} neighboring light curves. This might take a while."
    )
    lcfCol = search.download_all(flux_column=flux_column)

    # Pre-process the neighboring light curves
    lc_neighborhood = []
    lc_neighborhood_flux = []
    # Extract SAP light curves
    # We want zero-centered median normalized light curves
    for lc in lcfCol:
        lcSAP = lc.remove_nans().normalize()
        lcSAP.flux -= 1.0
        # Align or interpolate the neighboring target with the target under study
        if interpolate:
            # Interpolate to corrected_lc cadence times
            fInterp = PchipInterpolator(
                lcSAP.time.value,
                lcSAP.flux.value,
                extrapolate=extrapolate,
            )
            lc_neighborhood_flux.append(fInterp(corrected_lc.time.value))
        else:
            # The CBVs were aligned so also align the neighboring
            # lightcurves
            lc_trim_mask = np.in1d(
                lcSAP.cadenceno, corrected_lc.cadenceno
            )
            # If there are no non-trimmed cadences then nothing to add
            if (np.all(np.logical_not(lc_trim_mask))):
                continue
            lc_neighborhood_flux.append(lcSAP[lc_trim_mask].flux.value)
        lc_neighborhood.append(lcSAP)

    if len(lc_neighborhood) < min_targets:
        raise MinTargetsError(
            f"Unable to find at least {min_targets} neighbors within {radius} arcseconds radius."
        )
    # Store the unmolested lightcurve neighborhood but also save the
    # aligned or interpolated neighborhood flux
    from .. import LightCurveCollection  # local import to avoid circular import

    lc_neighborhood = LightCurveCollection(lc_neighborhood)
    lc_neighborhood_flux = lc_neighborhood_flux

    return lc_neighborhood, lc_neighborhood_flux


def _compute_correlation(fluxMatrix):
    """Finds the empirical target to target flux time series Pearson correlation.

    Parameters
    ----------
    fluxMatrix : float 2-d array[ntargets,ncadences]
        The matrix of target flux. There should be no gaps or Nans

    Returns
    -------
    correlation_matrix : [float 2-d array] (nTargets x nTargets)
        The target-target correlation
    """

    nCadences = len(fluxMatrix[:, 0])

    # Scale each flux value by the RMS flux for the given target.
    rmsFlux = np.sqrt(np.sum(fluxMatrix ** 2.0, axis=0) / nCadences)
    # If RMS is zero then set to Inf so that we don't get a divide by zero warning
    rmsFlux[np.nonzero(rmsFlux == 0.0)[0]] = np.inf
    unitNormFlux = fluxMatrix / np.tile(rmsFlux, (nCadences, 1))

    correlation_matrix = unitNormFlux.T.dot(unitNormFlux) / nCadences

    return correlation_matrix



"""DEPRECATED: `LightCurveFile` classes were removed in Lightkurve v2.0 in
favor of only having `LightCurve` classes.  To minimize breaking code, we
retain the `LightCurveFile` classes here as wrappers around the new
`LightCurve` objects, but will remove these wrappers in a future version..
"""
from . import LightCurve, KeplerLightCurve, TessLightCurve

LightCurveFile = LightCurve
KeplerLightCurveFile = KeplerLightCurve.read
TessLightCurveFile = TessLightCurve.read



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['lightkurve',
 'lightkurve.correctors',
 'lightkurve.io',
 'lightkurve.prf',
 'lightkurve.seismology']

package_data = \
{'': ['*'], 'lightkurve': ['data/*']}

install_requires = \
['astropy>=4.1',
 'astroquery>=0.3.10',
 'beautifulsoup4>=4.6.0',
 'bokeh>=1.0',
 'fbpca>=1.0',
 'matplotlib>=1.5.3',
 'numpy>=1.11',
 'oktopus>=0.1.2',
 'pandas>=1.1.4',
 'patsy>=0.5.0',
 'requests>=2.22.0',
 'scikit-learn>=0.24.0',
 'scipy>=0.19.0',
 'tqdm>=4.25.0',
 'uncertainties>=3.1.4']

extras_require = \
{':python_version >= "3.6" and python_version < "4.0"': ['memoization>=0.3.1']}

setup_kwargs = {
    'name': 'lightkurve',
    'version': '2.0.10',
    'description': 'A friendly package for Kepler & TESS time series analysis in Python.',
    'long_description': "Lightkurve\n==========\n\n**A friendly package for Kepler & TESS time series analysis in Python.**\n\n**Documentation: https://docs.lightkurve.org**\n\n|test-badge| |conda-badge| |pypi-badge| |pypi-downloads| |doi-badge| |astropy-badge|\n\n.. |conda-badge| image:: https://img.shields.io/conda/vn/conda-forge/lightkurve.svg\n                 :target: https://anaconda.org/conda-forge/lightkurve\n.. |pypi-badge| image:: https://img.shields.io/pypi/v/lightkurve.svg\n                :target: https://pypi.python.org/pypi/lightkurve\n.. |pypi-downloads| image:: https://pepy.tech/badge/lightkurve/month\n                :target: https://pepy.tech/project/lightkurve/month\n.. |test-badge| image:: https://github.com/lightkurve/lightkurve/workflows/Lightkurve-tests/badge.svg\n                 :target: https://github.com/lightkurve/lightkurve/actions?query=branch%3Amain\n.. |astropy-badge| image:: https://img.shields.io/badge/powered%20by-AstroPy-orange.svg?style=flat\n                   :target: http://www.astropy.org\n.. |doi-badge| image:: https://zenodo.org/badge/DOI/10.5281/zenodo.1181928.svg\n              :target: https://docs.lightkurve.org/about/citing.html             \n\n**Lightkurve** is a community-developed, open-source Python package which offers a beautiful and user-friendly way\nto analyze astronomical flux time series data,\nin particular the pixels and lightcurves obtained by\n**NASA's Kepler and TESS exoplanet missions**.\n\n.. image:: https://raw.githubusercontent.com/lightkurve/lightkurve/main/docs/source/_static/images/lightkurve-teaser.gif\n\nThis package aims to lower the barrier for students, astronomers,\nand citizen scientists interested in analyzing Kepler and TESS space telescope data.\nIt does this by providing **high-quality building blocks and tutorials**\nwhich enable both hand-tailored data analyses and advanced automated pipelines.\n\n\nDocumentation\n-------------\n\nRead the documentation at `https://docs.lightkurve.org <https://docs.lightkurve.org>`_.\n\n\nQuickstart\n----------\n\nPlease visit our quickstart guide at `https://docs.lightkurve.org/quickstart.html <https://docs.lightkurve.org/quickstart.html>`_.\n\n\nContributing\n------------\n\nWe welcome community contributions!\nPlease read the  guidelines at `https://docs.lightkurve.org/about/contributing.html <https://docs.lightkurve.org/about/contributing.html>`_.\n\n\nCiting\n------\n\nIf you find Lightkurve useful in your research, please cite it and give us a GitHub star!\nPlease read the citation instructions at `https://docs.lightkurve.org/about/citing.html <https://docs.lightkurve.org/about/citing.html>`_.\n\n\nContact\n-------\nLightkurve is an open source community project created by `the authors <AUTHORS.rst>`_.\nThe best way to contact us is to `open an issue <https://github.com/lightkurve/lightkurve/issues/new>`_ or to e-mail tesshelp@bigbang.gsfc.nasa.gov.\nPlease include a self-contained example that fully demonstrates your problem or question.\n",
    'author': 'Geert Barentsen',
    'author_email': 'hello@geert.io',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://docs.lightkurve.org',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6.1',
}


setup(**setup_kwargs)
